# gendoc: ignore

from docplex.mp.engine import LocalEngine, FakeSuccessEngine, FakeFailEngine
from docplex.mp.docloud_engine import DOcloudEngine


class EngineFactory(object):
    """ A factory class that manages creation of solver instances.
    """
    _engine_types_by_key = {"local": LocalEngine,
                            "fake": FakeSuccessEngine,
                            "fail": FakeFailEngine,
                            "docloud": DOcloudEngine}

    cplex_engine_type = None

    @classmethod
    def _get_engine_by_code(cls, code, default_engine):
        if code is None:
            return default_engine
        else:
            return cls._engine_types_by_key.get(code.lower(), default_engine)

    @classmethod
    def new_engine(cls, spec, env, model, docloud_context=None):
        """
        returns a new engine instance from a key
        :param spec:
        :param model:
        :return
        """
        has_cplex = env.has_cplex
        error_handler = model.error_handler

        if has_cplex:
            if not cls.cplex_engine_type:
                from docplex.mp.cplex_engine import CplexEngine
                cls.cplex_engine_type = CplexEngine
                cls._engine_types_by_key["cplex"] = CplexEngine
        elif docloud_context:
            model.warning("CPLEX DLL not found, will solve on DOcloud")

        # ---
        # what logic do we have:
        if has_cplex:
            # default is CPLEX if we have it
            engine_type = cls._get_engine_by_code(spec, cls.cplex_engine_type)
        elif docloud_context:
            # default is docloud
            engine_type = cls._get_engine_by_code(spec, DOcloudEngine)
        else:
            # no CPLEX, no credentials!
            model.warning("CPLEX DLL not found and model has no DOcloud credentials."
                                  "Provide credentials at solve time")
            engine_type = LocalEngine

        if not engine_type:
            model.fatal("Internal error, cannot build engine from spec {}", spec)

        return engine_type(error_handler, docloud_context=docloud_context)
