import platform
import os
import sys

# noinspection PyPep8
class Environment(object):
    """ This class detects and holds all information regarding other modules of interest.

    It detects whether CPLEX, numpy, and matplotlib are installed.
    """
    def __init__(self, start_auto_configure=True):
        """
        __init__(self)
        """
        self._found_cplex = False
        self._cplex_version = ''
        self._cplex_location = None
        self._found_numpy = False
        self._numpy_version = None
        self._found_pandas = None
        self._found_matplotlib = False
        self._matplotlib_version = None
        self._python_version = platform.python_version()
        self._system = platform.system()
        self._bitness = platform.architecture()[0]
        self._is64bit = sys.maxsize > 2**32

        if start_auto_configure:
            self.auto_configure()

    @property
    def has_cplex(self):
        """ Checks whether the CPLEX libraries are available.
        """
        return self._found_cplex

    @property
    def cplex_version(self):
        return self._cplex_version

    @property
    def has_matplotlib(self):
        """ Checks whether matplotlib libraries are available.
        """
        return self._found_matplotlib

    @property
    def has_pandas(self):
        """ Checks whether pandas libraries are available.
        """
        return self._found_pandas

    @property
    def cplex_location(self):
        """ Returns the system path where CPLEX is located, if present.
        """
        return self._cplex_location

    @property
    def has_numpy(self):
        """ Checks whether numpy libraries are available.
        """
        return self._found_numpy

    def is_64bit(self):
        """ Returns True if running on a 64-bit platform.
        """
        return self._is64bit

    def auto_configure(self):

        # detecting CPLEX
        try:
            import cplex
            self._found_cplex = True
            cpx = cplex.Cplex()
            self._cplex_version = cpx.get_version()
            cplex_module_file  = cplex.__file__
            if cplex_module_file:
                self._cplex_location = os.path.dirname(os.path.dirname(cplex_module_file))
            del cpx
        except ImportError:
            self._found_cplex = False

        try:
            import numpy.version as npv
            self._found_numpy = True
            self._numpy_version = npv.version
        except ImportError:
            self._found_numpy = False

        try:
            from matplotlib import __version__ as matplotlib_version
            self._found_matplotlib = True
            self._matplotlib_version = matplotlib_version
        except ImportError:
            self._found_matplotlib = False


        try:
            import pandas
            self._found_pandas = True
        except ImportError:
            self._found_pandas = False

        self.check_bitness_compatibility()

    def check_bitness_compatibility(self):
        if self._bitness and self._cplex_location:
            ok_bitness, ko_bitness = ("64", "32") if self._is64bit else ("32", "64")
            cplex_patform = os.path.basename(self._cplex_location)
            if cplex_patform.find(ok_bitness) >= 0:
                # good bitness is explicit in cplex location: sure OK
                pass
            elif cplex_patform.find(ko_bitness):
                # bad bitness is explicit: sure KO
                print("!!!! bitness inconsistency, python: {0}, CPLEX: {1}".format(self._bitness, cplex_patform))
            else:
                # cannot tell as cplex platform does not matche either bitness string,  maybe site package" ??
                pass

    @staticmethod
    def _display_feature(is_present, feature_name, feature_version, location=None):
        safe_feature_version = feature_version or "?"
        if is_present:
            if location:
                print("* {0} is present, version is {1}, located at: {2}".format(feature_name, safe_feature_version, location))
            else:
                print("* {0} is present, version is {1}".format(feature_name, safe_feature_version))
        else:
            print("* {0} is not available".format(feature_name))

    @property
    def max_nb_digits(self):
        return 16 if self.is_64bit() else 7

    @property
    def bitness(self):
        return 64 if self.is_64bit() else 32

    def print_information(self):
        print("* system is: {0} {1}".format(self._system, self._bitness))
        self._display_feature(self._found_cplex, "CPLEX wrapper", self._cplex_version, self._cplex_location)
        self._display_feature(self._found_numpy, "Numpy", self._numpy_version)
        self._display_feature(self._found_matplotlib, "Matplotlib", self._matplotlib_version)


    @staticmethod
    def closed_env():
        return Environment(start_auto_configure=False)

    @staticmethod
    def new_env(configure=True):
        return Environment(start_auto_configure=configure)

GlobalEnvironment = None

# This instance assumes nothing is found, CPLEX, numpy, etc, to be used for tests
ClosedEnvironment = Environment(start_auto_configure=False)

def get_no_graphics_env():
    # force matplotlib absent.
    env = Environment(start_auto_configure=True)
    # force matplotlib absent
    env._found_matplotlib = False
    env._found_numpy = False
    return env


def get_configured_env(display_env=False):
    global GlobalEnvironment
    if GlobalEnvironment is None:
        GlobalEnvironment = Environment(start_auto_configure=True)
        if display_env:
            GlobalEnvironment.print_information()
    return GlobalEnvironment


if __name__ == "__main__":
    GlobalEnvironment.print_information()
