# gendoc: ignore


class ProgressData(object):
    """ Singleton class to store progress info
    """

    def __init__(self):
        BIGNUM = 1e+75
        self.is_incumbent = False
        self.current_objective = BIGNUM
        self.best_bound = BIGNUM
        self.mip_gap = BIGNUM
        self.current_nb_nodes = 0
        self.remaining_nb_nodes = 0
        self.wall_time = 0.0
        self.det_time = 0.0

    def get_tuple(self):
        return (self.is_incumbent,
                self.current_objective,
                self.best_bound,
                self.mip_gap,
                self.current_nb_nodes,
                self.remaining_nb_nodes,
                self.wall_time,
                self.det_time)


class ProgressListener(object):

    def __init__(self):
        pass

    def notify_start(self):
        """ The method called when a solve has been initiated on a model
        """
        pass
    
    def notify_end(self, status, objective):
        """The method called when solve is finished on a model. The status is the solve status from the
        solve() method
        """
        pass

    def notify_progress(self, progress_data):
        pass


class TextProgressListener(ProgressListener):
    """ A simple implementation of Progress Listener, which prints messages to stdout
    """

    def __init__(self):
        ProgressListener.__init__(self)
        self._count = 0

    def formatted_gap(self, raw_gap):
        return "{:.2%}".format(raw_gap)

    def notify_progress(self, progress_data):
        self._count += 1
        current_obj = progress_data.current_objective
        best_bound  = progress_data.best_bound
        nb_nodes   = progress_data.current_nb_nodes
        remaining_nodes = progress_data.remaining_nb_nodes
        gap = self.formatted_gap(progress_data.mip_gap)

        print("{0:6}:  current={1}, best={2}, gap={3}, nodes={4}/{5}"
              .format(self._count, current_obj, best_bound, gap, nb_nodes, remaining_nodes))


class RecordProgressListener(ProgressListener):

    def __init__(self):
        ProgressListener.__init__(self)
        self.__recorded = []
        self.__final_objective = 1e+75
        self.__final_status = False

    def notify_start(self):
        # restart
        self.__recorded = []
        self.__final_objective = 1e+75

    def notify_progress(self, progress_data):
        self.__recorded.append(progress_data.get_tuple())

    def notify_end(self, status, objective):
        """ The method called when solve is finished on a model. The status is the solve status from the
        solve() method
        """
        self.__final_status = status
        if status:
            self.__final_objective = objective

    @property
    def final_objective(self):
        return self.__final_objective
    
    @property
    def final_status(self):
        return self.__final_status

    @property
    def size(self):
        return len(self.__recorded)

    def iter_progress_data(self):
        return iter(self.__recorded)
